#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) Markus Schmidt <schmidt@boomshop.net>
# License: F4A (free4all) - have a look in the LICENSE file that came with this
# software

import pygtk
pygtk.require('2.0')
import gtk 
import gobject
import pango
import os
import re
import sys
import mimetypes
import hashlib
import gobject
import cairo
import math
import signal
import shutil
import datetime
import time
import pyexiv2
import pyinotify
import urllib
import threading

from lib.Thumbnailer import Thumbnailer
from lib.Folder import Folder
from lib.Preview import Preview
from lib.Notifier import Notifier
from lib.Styles import Styles
from lib.Utils import Utils

import ConfigParser as config


gobject.threads_init()

glimpse = None

log_level = 1
def log(mesg, level = 0):
    if level <= log_level:
        print mesg

# TODO:
# self.set_scrollpos()
# create folders
# delete folders
# rename folders
# batch resize
# drag'n'drop
# graphical background

basedir = os.path.dirname(os.path.realpath(__file__))
if not os.path.exists(os.path.join(basedir, "glimpse.py")):
    if os.path.exists(os.path.join(os.getcwd(), "glimpse.py")):
        basedir = os.getcwd()
sys.path.insert(0, basedir)

class Glimpse(Utils):
    version = open(os.path.join(sys.path[0], 'version')).read().strip() # version number
    root     = "" # the directory to show
    userdir  = os.path.join(os.path.expanduser("~"), ".glimpse") # directory in users home
    thumbdir = os.path.join(userdir, "thumbs") # directory to store the thumbnails into
    basedir  = os.getcwd() # working directory
    progdir  = sys.path[0] # programs directory
    
    configfile = os.path.join(userdir, "glimpse.conf") # configuration file in users home
    default_settings = {
        "size": 1, # size id, 0 = small, 1 = mid, 2 = big
        "style": "dark", # selected style
        "editor_default": 0, # default editor application, id from list below
        "viewer_default": 0, # default viewer application, id from list below
        "autoclean": 365, # amount of days thumbnails are kept before they are deleted from disk on program start
        "language": "", # selected language
        "root": "", # selected directory
        "win_x": max(0, (gtk.gdk.screen_width() - 940) / 2), # horizontal window position
        "win_y": max(0, (gtk.gdk.screen_height() - 700) / 2), # vertical window position
        "win_w": 940, # horizontal window size
        "win_h": 700, # verical window size
        "hpaned": 260, # position of the horizontal panedwindow handle
        "vpaned": 420, # position of the vertical panedwindow handle
        "sidebar": 0, # sidebar visibility
        "infobar": 1, # infobar visibility
        "preview": 1, # preview visibility
        "font_factor": 0.66, # font scale factor, 1 keeps font at the same size in every viewport size setting
        "preview_border": 0, # border around the preview image
        "cache": 0, # cache thumbnails, uses more ram but is faster
    } # default settings to fork a new config from
    
    sizes = [
        {"name": "small", "size": 100,  "show": ["name"], "image": None},
        {"name": "mid",   "size": 150, "show": ["name", "sizeh", "resolutionh"], "image": None},
        {"name": "big",   "size": 200, "show": ["name", "sizeh", "resolutionh", "parent"], "image": None}
    ] # available sizes with names and other informations
    
    editors = [
        ["GIMP",        "gimp",    "gimp %s",    True],
        ["Fotoxx",      "fotoxx",  "fotoxx %s",  False],
        ["mtPaint",     "mtpaint", "mtpaint %s", False],
        ["ImageMagick", "display", "display %s", False],
    ] # all available editors to check for. [displayed name, command to which for, command to start, multi or single image command
    
    viewers = [
        ["Eye of GNOME",       "eog",       "eog %s",       False],
        ["ImageMagick",        "display",   "display %s",   False],
        ["Shotwell",           "shotwell",  "shotwell %s",  False],
        ["F-Spot",             "fspot",     "fspot  %s",    False],
        ["Geeqie",             "geeqie",    "geeqie %s",    False],
        ["GPicview",           "gpicview",  "gpicview %s",  False],
        ["gThumb",             "gthumb",    "gthumb %s",    False],
        ["Gwenview",           "gwenview",  "gewnview %s",  False],
        ["Mirage",             "mirage",    "mirage %s",    False],
        ["PhotoQT",            "photoqt",   "photoqt %s",   False],
        ["Quick Image Viewer", "qiv",       "qiv %s",       False],
        ["Ristretto",          "ristretto", "ristretto %s", True]
    ] # all available image viewers to check for
    
    mailers = [
        [ "Icedove",    "icedove",     "icedove --compose \"attachment='%s'\"",     True ],
        [ "Tunderbird", "thunderbird", "thunderbird --compose \"attachment='%s'\"", True ],
    ]
    
    size = 1 # CONFIG - the default size (just a preset)
    style = "dark" # CONFIG - the actual style (just a preset)
    editor_default = 1 # CONFIG
    viewer_default = 0 # CONFIG
    
    infos = ["number", "name", "type", "path", "resolutionh", "sizeh", "size", "hash", "md5"] # infos avilable in the info textview
    info = None # image displayed in the preview
    
    folders = [] # array containing all folders with some images below root
    active = [] # array containing all selected image objects
    last_active = [] # copy of the active array to redraw old folder surfaces when selection changed
    
    actual = None # the last clicked image object
    hover = None # the hovered image object
    last_hover = None # keep the last hovered image to reduce redrawing
    lastmove = None # last event object of a mouse move to trigger it again later when things change
    
    _loadingimg = "loading.png" # the loading image located in progdir
    _win = [0,0] # workaround to remember latest window size to determine if things changed
    
    language = "en" # CONFIG - default language if everything else fails
    languages = {} # dict containing all available languages after __init__
    
    lock  = threading.Lock() # lock folders on manipulation for threading-safe execution
    plock = threading.Lock() # lock the preview image for threading-safe execution
    clock = threading.Lock() # lock for the counter for threading-safe execution
    wlock = threading.Lock() # lock for watch manager
    
    orient_icons = [] # aray containing all orientation icons for the actual selected theme. is filled on theme change
    options = None # options gtkwindow. is remembered for toggling on and off via toolbar button
    helpwin = None # help gtkwindow. is remembered for toggling on and off via toolbar button
    previewimg = None # preview pixbuf. is generated via Preview thread.
    previewgen = None # Preview generator. Is remembered to check if one is running.
    up = sys.maxint / 2 # upsample hack. is used to generate sortable ids from fid and iid for all images in the selection which can be decoded to fid and iid again
    destroying = False # if we want to destroy the window (and glimpse in general) it is used to stop running threads and prevent them from respawning
    scrollpos = 0
    reading_dir = False
    notifier = None # pyinotify object
    watchmask = pyinotify.IN_DELETE | pyinotify.IN_CREATE | pyinotify.IN_MODIFY | pyinotify.IN_MOVED_TO | pyinotify.IN_MOVED_FROM
    watch = {}
    _active = [] # list of selected path names for faster checks, is maintained by append and remove functions
    mainpaned_pos = 0 # position of the vertical paned window separator in pixels from bottom
    mainpaned_height = 0 # height of the paned window
    queuelen = 0
    loaded = 0
    
    def __init__(self):
        #signal.signal(signal.SIGINT, signal.SIG_DFL)
        signal.signal(signal.SIGINT, self.ctrl_c)
        
        # user and thumbnail directory
        if not os.path.exists(self.userdir):
            os.mkdir(self.userdir)
        if not os.path.exists(self.thumbdir):
            os.mkdir(self.thumbdir)
        
        # config
        self.config = self.start_config(self.configfile,
                                        self.default_settings,
                                        ["glimpse"],
                                        True)
        
        # thumbnailer
        self.thumbnailer = Thumbnailer(self.counter_cb)
        self.thumbnailer.set_size(self.sizes[len(self.sizes) - 1]["size"],
                                  self.sizes[len(self.sizes) - 1]["size"])
        
        # loading thumbs
        for s in self.sizes:
            s["image"] = cairo.ImageSurface.create_from_png(os.path.join(self.progdir, self._loadingimg))
            
        # Initiate Language
        gtklang = str(gtk.get_default_language()).replace("-", "_").lower()[0:5]
        fallb = None
        found = False
        for root, dirs, files in os.walk(os.path.join(sys.path[0], "languages")):
            for f in files:
                name = str(f)[:-3]
                if not re.search(r"^[a-zA-Z]{2}(_[a-zA-Z]{2}){0,1}$",
                                 name): continue
                tmp = __import__("languages.%s" %(name), globals(), locals(),
                                ["locale", "language"], -1)
                self.languages[tmp.locale] = tmp.language
                if gtklang == name.lower():
                    self.language = name
                    found = True
                elif not fallb and gtklang[0:2] == name.lower()[0:2]:
                    fallb = name
        if not found and fallb:
            self.language = fallb
        if (self.config.get("glimpse", "language")
            and self.config.get("glimpse", "language") != ""
            and self.config.get("glimpse", "language") in self.languages):
            self.language = self.config.get("glimpse", "language")
        
        if self.config.get("glimpse", "language") != self.language:
            self.config.set("glimpse", "language", self.language)
            self.write_config()
        
        self.set_language()
        
        # size
        if self.config.has_option("glimpse", "size"):
            self.size = int(self.config.get("glimpse", "size"))
            
        # style
        if self.config.has_option("glimpse", "style"):
            self.style = self.config.get("glimpse", "style")
        
        # viewer_default
        if self.config.has_option("glimpse", "viewer_default"):
            self.viewer_default = int(self.config.get("glimpse", "viewer_default"))
            
        # editor_default
        if self.config.has_option("glimpse", "editor_default"):
            self.editor_default = int(self.config.get("glimpse", "editor_default"))
        
        # build logo
        self.logo = gtk.Image()
        self.logo.set_from_file(os.path.join(self.progdir, "Logo.png"))
        
        ##############################################
        
        # build window
        self.window = gtk.Window(gtk.WINDOW_TOPLEVEL)
        self.window.set_title("glimpse " + self.version)
        self.window.connect("destroy", self.destroy)
        self.window.connect("delete_event", self.delete_event)
        self.window.set_icon_from_file(os.path.join(self.progdir, "glimpse.png"))
        
        # build box
        self.box = gtk.VBox(False, 0)
        self.window.add(self.box)
        self.box.set_border_width(10)
        
        ##############################################
        
        # build toolbar
        self.toolbar = gtk.Toolbar()
        self.toolbar.set_property("toolbar-style", gtk.TOOLBAR_BOTH)
        self.box.pack_start(self.toolbar, False, False, 0)
        
        # filechooserbutton
        ti = gtk.ToolItem()
        a = gtk.Alignment(0.5, 0.5, 0, 0)
        self.filechooserbutton = gtk.FileChooserButton("")
        self.filechooserbutton.set_action(gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER)
        ti.add(a)
        a.add(self.filechooserbutton)
        self.toolbar.insert(ti, -1)
        
        # loadbutton
        self.loadbutton = gtk.ToolButton(gtk.STOCK_OPEN)
        self.toolbar.insert(self.loadbutton, -1)
        self.loadbutton.connect("clicked", self.loadbutton_cb)
        
        # separator
        ex = gtk.SeparatorToolItem()
        ex.set_draw(False)
        self.toolbar.insert(ex, -1)
        self.toolbar.child_set_property(ex, "expand", True)
        
        # style combobox
        ti = gtk.ToolItem()
        a = gtk.Alignment(0.5, 0.5, 0, 0)
        self.stylebox = gtk.combo_box_new_text()
        ti.add(a)
        a.add(self.stylebox)
        self.toolbar.insert(ti, -1)
        
        self.stylebox.connect('changed', self.stylebox_cb)
        
        # separator
        ex = gtk.SeparatorToolItem()
        ex.set_draw(False)
        self.toolbar.insert(ex, -1)
        
        # size combobox
        ti = gtk.ToolItem()
        a = gtk.Alignment(0.5, 0.5, 0, 0)
        self.sizebox = gtk.combo_box_new_text()
        ti.add(a)
        a.add(self.sizebox)
        self.toolbar.insert(ti, -1)
        
        self.sizebox.connect('changed', self.sizebox_cb)
        
        # separator
        ex = gtk.SeparatorToolItem()
        ex.set_draw(False)
        self.toolbar.insert(ex, -1)
        self.toolbar.child_set_property(ex, "expand", True)
        
        # sidebar button
        self.sidebarbutton = gtk.ToolButton(gtk.STOCK_DIRECTORY)
        self.toolbar.insert(self.sidebarbutton, -1)
        self.sidebarbutton.connect("clicked", self.sidebarbutton_cb)
        
        # info button
        self.infobutton = gtk.ToolButton(gtk.STOCK_INFO)
        self.toolbar.insert(self.infobutton, -1)
        self.infobutton.connect("clicked", self.infobutton_cb)
        
        # optionsbutton
        self.optionsbutton = gtk.ToolButton(gtk.STOCK_PREFERENCES)
        self.toolbar.insert(self.optionsbutton, -1)
        self.optionsbutton.connect("clicked", self.optionsbutton_cb)
        
        # helpbutton
        self.helpbutton = gtk.ToolButton(gtk.STOCK_HELP)
        self.toolbar.insert(self.helpbutton, -1)
        self.helpbutton.connect("clicked", self.help_cb)
        
        ##############################################
        
        # build mainpaned
        self.mainpaned = gtk.VPaned()
        self.mainpaned.set_position(int(self.config.get("glimpse", "vpaned")))
        self.box.pack_start(self.mainpaned, True, True, 0)
        self.mainpaned.connect_after("notify", self.mainpaned_cb)
        
        # build toppaned
        self.toppaned = gtk.HPaned()
        self.toppaned.set_position(int(self.config.get("glimpse", "hpaned")))
        self.mainpaned.add1(self.toppaned)
        self.toppaned.connect_after("notify", self.toppaned_cb)
        
        # build bottompaned
        self.bottompaned = gtk.HPaned()
        self.bottompaned.set_position(int(self.config.get("glimpse", "hpaned")))
        self.mainpaned.add2(self.bottompaned)
        self.bottompaned.connect_after("notify", self.bottompaned_cb)
        
        ################################################
        
        # build filechooser
        self.filechooser = gtk.FileChooserWidget(gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER)
        self.filechooser.get_children()[0].get_children()[0].get_children()[1].get_children()[0].hide()
        tview = self.filechooser.get_children()[0].get_children()[0].get_children()[1].get_children()[1].get_children()[0].get_children()[0].get_children()[0]
        self.toppaned.add1(self.filechooser)
        self.filechooser.connect("selection-changed", self.filechooser_cb)
        c = 0
        for col in tview.get_columns():
            if c != 0:
                col.set_visible(False)
            c += 1
        
        # build preview
        self.preview = gtk.DrawingArea()
        self.bottompaned.add1(self.preview)
        self.preview.connect("expose-event", self.expose_preview)
        
        # build infoscroller
        self.infoscroll = gtk.ScrolledWindow()
        self.infoscroll.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        self.bottompaned.add2(self.infoscroll)
        
        # build infotext
        self.infoview = gtk.TextView()
        self.infoview.set_editable(False)
        self.infoview.modify_font(pango.FontDescription('monospace'))
        self.infoview.set_border_width(10)
        self.infoscroll.add(self.infoview)
        
        ################################################
        
        # bild eventbox
        self.eventbox = gtk.EventBox()
        self.toppaned.add2(self.eventbox)
        
        # build scroller
        self.scroll = gtk.ScrolledWindow()
        self.scroll.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_ALWAYS)
        self.eventbox.add(self.scroll)
        self.scrolladjust = self.scroll.get_vadjustment()
        self.scrolladjust.connect_after("value-changed", self.scroll_cb)
        
        # build drawing area container
        self.drawings = gtk.VBox()
        self.scroll.add_with_viewport(self.drawings)
        self.viewport = self.scroll.get_children()[0]
        self.viewport.set_shadow_type(gtk.SHADOW_NONE)
        self.vmargin = (self.viewport.style_get_property("focus-line-width")
                    + self.viewport.style_get_property("focus-padding")) * 2
        
        # progress bar
        self.progress = gtk.ProgressBar()
        self.progress.set_text("")
        self.progress.set_fraction(0)
        self.progress.set_pulse_step(0.05)
        self.box.pack_start(self.progress, False, False, 0)
        
        # cursor
        self.cursor = gtk.gdk.Cursor(gtk.gdk.WATCH)
        
        # root folder
        if self.config.get("glimpse", "root"):
            self.root = self.config.get("glimpse", "root")
        try:
            root = sys.argv[len(sys.argv) - 1]
        except:
            root = ""
        if root and os.path.isdir(root):
            self.root = root
        if self.root and self.root != os.path.expanduser("~"):
            while not os.path.isdir(self.root):
                self.root = os.path.basename(self.root)[0]
        else:
            fc = gtk.FileChooserDialog(self.lang.filechooser_folder,
                                       self.window,
                                       gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER,
                                       (self.lang.move_ok, gtk.RESPONSE_OK))
            fc.set_current_folder(os.path.expanduser("~"))
            fc.show_all()
            response = fc.run()
            self.root = fc.get_current_folder()
            fc.destroy()
        
        if self.config.get("glimpse", "root") != self.root:
            self.config.set("glimpse", "root", self.root)
            self.write_config()
        self.filechooser.set_current_folder(self.root)
        self.filechooserbutton.set_current_folder(self.root)
        
        # cleanup
        self.cleanup_thumbs(int(float(self.config.get("glimpse", "autoclean"))))
        
        self.fill_language()
        
        # window position and size
        self.window.set_default_size(int(self.config.get("glimpse", "win_w")),
                                     int(self.config.get("glimpse", "win_h")))
        self.window.show_all()
        self.window.move(int(self.config.get("glimpse", "win_x")),
                         int(self.config.get("glimpse", "win_y")))
        
        # hide/show stuff
        self.show_infobar(int(self.config.get("glimpse", "infobar")))
        self.show_sidebar(int(self.config.get("glimpse", "sidebar")))
        self.show_preview(int(self.config.get("glimpse", "preview")))
        
        # style
        self.set_style(self.style)
        
        self.force()
        
        self.mainpaned_pos = self.mainpaned.get_allocation().height - self.mainpaned.get_position()
        
        # events after everything is displayed
        self.preview.add_events(gtk.gdk.BUTTON_PRESS_MASK
                              | gtk.gdk.BUTTON_RELEASE_MASK)
        self.preview.connect("button_press_event", self.preview_click_cb)
        self.eventbox.add_events(gtk.gdk.STRUCTURE_MASK)
        self.eventbox.connect('expose-event', self.resize_cb)
        
        self.window.connect("check-resize", self.window_resize_cb)
        
        # file system watch manager
        self.watchmanager = pyinotify.WatchManager()
        self.notifier = pyinotify.ThreadedNotifier(self.watchmanager, Notifier(self))
        self.notifier.start()
        self.watch = {}
        
        # finally...
        self.load(self.root)
        
        # focus
        focus = self.eventbox
        focus.add_events(gtk.gdk.BUTTON_PRESS_MASK)
        focus.connect('key_press_event', self.key_press_cb)
        focus.connect("button_press_event", self.grab_focus)
        focus.set_can_focus(True)
        focus.grab_focus() 
        
    #################
    # FUNCTIONALITY #
    #################
    
    def show_preview(self, p):
        # hide or show the preview image
        if p:
            self.preview.show()
        else:
            self.preview.hide()
            
    def show_sidebar(self, p):
        # hide or show the sidebar
        if p:
            self.filechooser.show()
            self.filechooserbutton.hide()
        else:
            self.filechooser.hide()
            self.filechooserbutton.show()
        self.draw_folders()
            
    def show_infobar(self, p):
        # hide or show the info bar
        if p:
            self.bottompaned.show()
        else:
            self.bottompaned.hide()
            
    def set_size(self, size):
        # set a size. propagate it to every folder
        self.size = min(len(self.sizes) - 1, max(0, size))
        self.config.set("glimpse", "size", self.size)
        self.lock.acquire()
        try:
            for f in self.folders:
                f.set_size(self.sizes[self.size])
        finally:
            self.lock.release()
        
    def set_style(self, style):
        # set a style. includes setting all gtk styles of all containers
        # beneath the folder surfaces and to propagate it to all
        # folders. additionally the orientation icons are read as cairo
        # surfaces
        
        self.style = style
        self.config.set("glimpse", "style", self.style)
        
        l = [self.drawings, self.scroll, self.viewport, self.infoview]
        for t in l:
            self.style_widget(t)
        self.lock.acquire()
        try:
            for f in self.folders:
                self.style_widget(f.widget)
                
            # orient icons
            self.orient_icons = []
            for i in range(2, 9):
                self.orient_icons.append(
                    cairo.ImageSurface.create_from_png(os.path.join(self.progdir, "icons", Styles[self.style]["icons"], "orient_%d.png" % i)))
                    
            self.preview.queue_draw()
        
            for f in self.folders:
                f.set_style(self.style)
        finally:
            self.lock.release()
        
    def load(self, directory):
        self.reading_dir = True
        self.lock.acquire()
        try:
            self.folders = []
            self.active = []
            self.hover = None
            self.thumbnails = []
            self.thumbnailer.clear()
            self.queuelen = 0
            self.loaded = 0
        finally:
            self.lock.release()
        self.drawings.foreach(self.drawings.remove)
        self.drawings.pack_start(self.logo, True, True, 0)
        self.logo.set_size_request(self.viewport.get_allocation().width, self.viewport.get_allocation().height)
        self.logo.show()
        self.progress.show()
        self.force()
        self.wlock.acquire()
        try:
            if len(self.watch):
                self.watchmanager.rm_watch(self.watch.values())
                self.watch = {}
        finally:
            self.wlock.release()
        self.read_dir(directory, 0, False)
        self.drawings.remove(self.logo)
        self.progress.hide()
        self.force()
        self.reading_dir = False
    
    def rename_file(self, frm, to):
        try:
            os.rename(frm, to)
        except:
            log("Unable to rename %s to %s" % (frm, to), 0)
    
    def delete_file(self, f):
        try:
            os.unlink(f)
        except:
            log("Unable to delete %s" % (f), 0)
    
    def copy_file(self, frm, to):
        try:
            shutil.copy(frm, to)
        except:
            log("Unable to copy %s to %s" % (frm, to), 0)
            
    def move_file(self, image, targ):
        try:
            shutil.move(image, targ)
        except:
            log("Unable to move %s to %s" % (image, targ), 0)
    
    def show_info(self, img):
        self.info = img
        txt = self.build_info(img, False)
        ex = self.build_exif(img)
        if ex:
            txt += self.lang.info_exif + ex.replace('\x00', '').decode('utf-8', 'replace').encode('utf-8')
        self.infoview.get_buffer().set_text(txt)
        self.request_preview(img)
    
    def show_info_multi(self, lst):
        self.info = self.actual
        txt = ""
        b = 0
        c = 1
        fs = []
        self.lock.acquire()
        try:
            for i in lst:
                txt += self.build_info(i, True, c) + "\n"
                b += i.size
                if not i.folder.id in fs:
                    fs.append(i.folder.id)
                c += 1
            f = (self.lang.info_folders % len(fs)) if len(fs) > 1 else ""
            txt = self.lang.info_multi % (len(lst), f, self.bytes2string(b) + "Bytes", txt)
            self.infoview.get_buffer().set_text(txt)
            self.request_preview(self.hover)
        finally:
            self.lock.release()
            
    def cleanup_thumbs(self, days):
        d = time.time() - days * 24 * 60 * 60
        for f in os.listdir(self.thumbdir):
            f = os.path.join(self.thumbdir, f)
            if os.path.getctime(f) < d:
                os.unlink(f)
                
    def clear_selection(self):
        self.infoview.get_buffer().set_text("")
        self.plock.acquire()
        try:
            self.previewimg = None
        finally:
            self.plock.release()
        self.preview.queue_draw()
        self.active_set([])
        
            
    ####################
    # WINDOW CALLBACKS #
    ####################
    
    def store_scrollpos(self):
        self.scrollpos = float(self.scrolladjust.get_value()) / self.scrolladjust.get_upper()
    
    def set_scrollpos(self):
        adj = float(self.scrolladjust.get_upper()) * self.scrollpos
        self.scrolladjust.set_value(adj)
    
    def scroll_cb(self, widget):
        self.store_scrollpos()
    
    def filechooser_cb(self, widget):
        f = self.has_folder(widget.get_filename())
        try:
            a = self.scroll.get_vadjustment()
            a.set_value(min(a.get_upper() - widget.allocation.height, f.widget.allocation.y))
        except:
            return
    
    def mainpaned_cb(self, widget, gparam):
        if gparam.name == 'position':
            n = float(widget.get_position()) / self.mainpaned.get_allocation().height
            if n < 0.95:
                self.mainpaned_pos = widget.get_allocation().height - widget.get_position()
            
    def toppaned_cb(self, widget, gparam):
        if gparam.name == 'position':
            self.bottompaned.set_position(widget.get_position())
            self.lock.acquire()
            try:
                for r in self.folders:
                    r.resize()
            finally:
                self.lock.release()
        
    def bottompaned_cb(self, widget, gparam):
        if gparam.name == 'position':
            self.toppaned.set_position(widget.get_position())
    
    def help_cb(self, widget):
        if self.helpwin:
            self.helpwin.destroy()
            self.helpwin = None
        else:
            self.helpwin = self.help()
    
    def key_press_cb(self, widget, event):
        key = self.key_event_to_string(event)
        
        if key == "CTRL+PLUS":
            self.set_size(self.size + 1)
            return
        if key == "CTRL+MINUS":
            self.set_size(self.size - 1)
            return
        if key == "CTRL+T":
            style = Styles.keys().index(self.style)
            self.set_style(Styles.items()[(style + 1) % len(Styles)][0])
            return
            
        if not len(self.active):
            return
        
        reshov = False
        if not self.hover:
            self.hover = self.active[len(self.active) - 1]
            reshov = True
            
        if key == "CTRL+V" or key == "RETURN":
            self.execute_cb(widget, self.viewers[self.viewer_default])
        if key == "CTRL+E":
            self.execute_cb(widget, self.editors[self.editor_default])
        if key == "CTRL+M":
            self.move_cb(widget)
        if key == "CTRL+C":
            self.copy_cb(widget)
        if key == "CTRL+R":
            self.rename_cb(widget)
        if key == "DELETE":
            self.delete_cb(widget)
            
        if reshov:
            self.hover = None
   
    
        
    def resize_cb(self, widget, event):
        if self.scroll.allocation.width != self._win[0] or self.scroll.allocation.height != self._win[1]:
            self.lock.acquire()
            try:
                for r in self.folders:
                    r.resize()
            finally:
                self.lock.release()
            self._win = (
                self.scroll.allocation.width,
                self.scroll.allocation.height
            )
            
    def window_resize_cb(self, widget):
        h = self.mainpaned.get_allocation().height
        if h != self.mainpaned_height:
            self.mainpaned.set_position(int(h - self.mainpaned_pos))
            self.mainpaned_height = h
        
    def grab_focus(self, widget, event):
        widget.grab_focus()
        
    ##################
    # MENU CALLBACKS #
    ##################
    
    def execute_cb(self, widget, *data):
        s = ""
        self.lock.acquire()
        try:
            p = data[0]
            for a in self.active:
                if p[3]:
                    s += "\"%s\" " % a.path
                else:
                    os.system((p[2] % ("\"%s\"" % a.path)) + " 2>&1 > /dev/null &")
            if p[3]:
                os.system((p[2] % s) + " 2>&1 > /dev/null &")
        finally:
            self.lock.release()
        # just a cheap hack to simulate responsivity:
        self.window.window.set_cursor(self.cursor)
        self.force()
        time.sleep(1)
        self.window.window.set_cursor(None)
    
    def folder_cb(self, widget):
        self.lock.acquire()
        try:
            os.system("xdg-open '" + self.actual.folder.path + "'")
        finally:
            self.lock.release()
    
    def send_to_cb(self, widget, mailer):
        self.lock.acquire()
        try:
            s = ""
            if mailer[3]:
                for a in self.active:
                    s += ("%s," % a.path)
                s = s[0:-1]
            else:
                s = self.active[0].path
            os.system(mailer[2] % s)
        finally:
            self.lock.release()
            
    def _rename_cb(self, exp):
        m = re.search("%([0-9]*)", exp)
        numf = ""
        nums = ""
        if m and m.group(1):
            nums = m.group(1)
            numf = "%0" + str(len(m.group(1))) + "d"
        return self.decode_expression(exp, self.actual, nums, numf, 0)
        
    def rename_cb(self, widget):
        if len(self.active) == 1:
            pre = self.active[0].name
        else:
            pre = "%n.%e"
        exp = self.entry_dialog(self.lang.rename_title,
                                self.lang.rename_entry,
                                self.lang.rename_text,
                                pre,
                                (0, pre.rfind('.')),
                                self.lang.expression_help,
                                self.lang.expression_title,
                                self._rename_cb)
        if exp:
            lm = len(self.active) - 1
            count = 0
            m = re.search("%([0-9]*)", exp)
            numf = ""
            nums = ""
            if m and m.group(1):
                nums = m.group(1)
                numf = "%0" + str(len(m.group(1))) + "d"
                
            self.active = sorted(self.active, key=lambda x: x.path.lower())
            multi = False
            for a in self.active:
                _n = self.decode_expression(exp, a, nums, numf, count)
                source = a.path
                target = os.path.join(a.parent, _n)
                if multi:
                    self.rename_file(source, target)
                elif os.path.exists(target) and target != source:
                    o = self.overwrite_dialog(self.lang.overwrite_title,
                                              self.lang.overwrite_text % _n,
                                              self.lang.overwrite_name,
                                              _n,
                                              (0, _n.rfind('.')),
                                              len(self.active) > 1)
                    if type(o) == type(''):
                        target = os.path.join(a.parent, o)
                    elif o == -1:
                        break
                    elif not o:
                        continue
                    elif o:
                        multi = True
                    self.rename_file(source, target)
                elif target != source:
                    self.rename_file(source, target)
                    
                count += 1
            
    def copy_cb(self, widget):
        if len(self.active) == 1:
            pre = self.lang.copy_fname_single % self.active[0].name
        else:
            pre = self.lang.copy_fname_multi
        exp = self.entry_dialog(self.lang.copy_title,
                                self.lang.copy_entry,
                                self.lang.copy_text,
                                pre,
                                (0, pre.rfind('.')),
                                self.lang.expression_help,
                                self.lang.expression_title,
                                self._rename_cb)
        if exp:
            self.lock.acquire()
            try:
                lc = 0
                lm = len(self.active) - 1
                count = 0
                m = re.search("%([0-9]*)", exp)
                numf = ""
                nums = ""
                if m and m.group(1):
                    nums = m.group(1)
                    numf = "%0" + str(len(m.group(1))) + "d"
                    
                self.active = sorted(self.active, key=lambda x: x.path.lower())
                multi = False
                for a in self.active:
                    _n = self.decode_expression(exp, a, nums, numf, count)
                    source = a.path
                    target = os.path.join(a.parent, _n)
                    if multi:
                        self.copy_file(source, target)
                    elif os.path.exists(target) and target != source:
                        o = self.overwrite_dialog(self.lang.overwrite_title,
                                                  self.lang.overwrite_text % _n,
                                                  self.lang.overwrite_name,
                                                  _n,
                                                  (0, _n.rfind('.')),
                                                  len(self.active) > 1)
                        if type(o) == type(''):
                            target = os.path.join(a.parent, o)
                        elif o == -1:
                            break
                        elif not o:
                            continue
                        elif o:
                            multi = True
                        self.copy_file(source, target)
                    elif target != source:
                        self.copy_file(source, target)
                        
                    lc += 1
                    count += 1
            finally:
                self.lock.release()
            
    def delete_cb(self, widget):
        f = self.actual.folder
        i = self.actual
        l = len(self.active)
        if l == 1:
            title = self.lang.delete_title
            text  = self.lang.delete_text_single % i.name
        else:
            title = self.lang.delete_title
            s = ""
            c = 0
            for j in self.active:
                if c < 5:
                    s += "%s\n" % j.name
                elif c == 5:
                    s += "...\n"
                else:
                    break
                c += 1
            text  = self.lang.delete_text_multi % (len(self.active), s)
        if self.dialog(title, text):
            self.lock.acquire()
            try:
                lc = 0
                lm = len(self.active) - 1
                for j in self.active:
                    if lc >= lm:
                        self.delete_file(j.path)
                        lc += 1
                self.active = []
                self.hover = None
                self.actual = None
            finally:
                self.lock.release()
                
                
    def move_cb(self, widget):
        fc = gtk.FileChooserDialog(self.lang.move_title,
                                   self.window,
                                   gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER,
                                   (self.lang.move_ok, gtk.RESPONSE_OK, self.lang.move_cancel, gtk.RESPONSE_CANCEL))
        fc.set_current_folder(self.actual.folder.path)
        fc.show_all()
        response = fc.run()
        
        targ = fc.get_current_folder()
            
        if response == gtk.RESPONSE_CANCEL or not os.path.isdir(targ):
            fc.destroy()
            return
        fc.destroy()
        
        tf = self.get_folder_by_path(targ)
        self.lock.acquire()
        try:
            multi = False
            lc = 0
            lm = len(self.active) - 1
            for j in self.active:
                if j.parent != targ:
                    _n = j.name
                    source = j.path
                    target = os.path.join(targ, _n)
                    if multi:
                        self.move_file(source, target)
                    elif os.path.exists(target) and target != source:
                        o = self.overwrite_dialog(self.lang.overwrite_title,
                                                  self.lang.overwrite_text % _n,
                                                  self.lang.overwrite_name,
                                                  _n,
                                                  (0, _n.rfind('.')),
                                                  len(self.active) > 1)
                        if type(o) == type(''):
                            target = os.path.join(targ, o)
                        elif o == -1:
                            break
                        elif not o:
                            continue
                        elif o:
                            multi = True
                        self.move_file(source, target)
                    elif target != source:
                        self.move_file(source, target)
                    
                lc += 1
            self.active = []
            self.hover = None
            self.actual = None
        finally:
            self.lock.release()
    
    ####################+
    # TOOLBAR CALLBACKS #
    #####################
    
    def loadbutton_cb(self, widget):
        if int(self.config.get("glimpse", "sidebar")):
            self.root = self.filechooser.get_current_folder()
            self.filechooserbutton.set_current_folder(self.root)
        else:
            self.root = self.filechooserbutton.get_current_folder()
            self.filechooser.set_current_folder(self.root)
        self.config.set("glimpse", "root", self.root)
        self.load(self.root)
        
    def optionsbutton_cb(self, widget):
        self.options_window()
    
    def sidebarbutton_cb(self, widget):
        s = int(not int(self.config.get("glimpse", "sidebar")))
        self.config.set("glimpse", "sidebar", s)
        self.show_sidebar(s)
        self.write_config()
        
    def infobutton_cb(self, widget):
        s = int(not int(self.config.get("glimpse", "infobar")))
        self.config.set("glimpse", "infobar", s)
        self.show_infobar(s)
        self.write_config()
        
    def sizebox_cb(self, widget):
        self.set_size(widget.get_active())
        
    def stylebox_cb(self, widget):
        self.set_style(Styles.items()[widget.get_active()][0])
    
    ##########################
    # DRAWING AREA CALLBACKS #
    ##########################
    
    def mousemove_cb(self, widget, event, f):
        self.hover = self.get_image(event.x, event.y, f)
        if self.hover != self.last_hover:
            if self.hover:
                widget.set_tooltip_text(self.hover.name)
                widget.set_has_tooltip(True)
            else:
                widget.set_has_tooltip(False)
            self.force()
            f.widget.queue_draw()
            if self.last_hover:
                self.last_hover.folder.widget.queue_draw()
            self.last_hover = self.hover
        self.lastmove = (widget, event.copy(), f)
        return True
        
    def drawing_click_cb(self, widget, *data):
        event = data[0]
        f = data[1]
        ap = False
        if self.hover:
            self.actual = self.hover
            # left click
            if event.button == 1:
                if event.type == gtk.gdk._2BUTTON_PRESS:
                    self.execute_cb(widget, self.viewers[self.viewer_default])
                    return
                # CTRL pressed
                if event.get_state() & gtk.gdk.CONTROL_MASK:
                    ap = False
                    for a in self.active:
                        if self.hover == a:
                            self.active_remove(a)
                            ap = True
                            break
                    if not ap:
                        self.active_append(self.hover)
                # SHIFT pressed
                elif event.get_state() & gtk.gdk.SHIFT_MASK:
                    _l = len(self.active)
                    if _l:
                        # get lowest an highest image id in self.active
                        highest = 0
                        lowest  = len(self.folders) * self.up
                        for a in self.active:
                            tmp = self.get_intid(a)
                            lowest = min(tmp, lowest)
                            highest = max(tmp, highest)
                        
                        # get actually clicked image id which is the
                        # target we want to select
                        targ  = self.get_intid(self.hover)
                        targf = self.hover.folder.id
                        targi = self.hover.id
                        
                        # determine the starting image and folder we
                        # want to mark from
                        # if some images are already marked we want to
                        # start from the upper or lower end of the actual
                        # selection
                        if targ > lowest and targ < highest:
                            start  = self.get_intid(self.active[_l - 1])
                        elif targ > highest:
                            start = lowest
                        else:
                            start = highest
                        
                        # set start ids
                        startf = self.get_folder_int(start)
                        starti = self.get_image_int(start)
                            
                        # stepsize
                        step = 1 if targ > start else -1
                        
                        act = start
                        self.active = []
                        
                        fs = []
                        
                        while True:
                            actf = self.get_folder_int(act)
                            acti = self.get_image_int(act)
                            self.active_append(self.folders[actf].images[acti])
                            if not actf in fs:
                                fs.append(actf)
                            
                            # check if we have to switch folders
                            if step > 0:
                                # stepping forward
                                l = len(self.folders[actf].images) - 1
                                if acti == l and actf < targf:
                                    # we reached last image in folder and
                                    # have at least one folder left to cycle
                                    # through - step one folder forward
                                    act += self.up - l
                                    continue
                            else:
                                # rolling backwards
                                if not acti and actf > targf:
                                    # we reached first image in folder and
                                    # have at least one folder left to cycle
                                    # through - reduce the internal folder id
                                    act -= self.up - len(self.folders[actf - 1].images) + 1
                                    continue
                            
                            # check if we have to end the cycle
                            if acti == targi and actf == targf:
                                # we reached target
                                break
                                
                            act += step
                        for i in fs:
                            self.folders[i].widget.queue_draw()
                    else:
                        self.active_set([self.hover])
                # normal behavior
                else:
                    self.active_set([self.hover])
            # right click
            elif event.button == 3:
                for a in self.active:
                    if self.hover == a:
                        ap = True
                        break
                if not ap:
                    self.active_set([self.hover])
                self.show_image_menu(event.x, event.y, self.hover)
            if len(self.active) > 1:
                self.show_info_multi(self.active)
            else:
                self.show_info(self.actual)
        else:
            # clicked outside any images
            self.clear_selection()
        f.widget.queue_draw()
        if len(self.last_active):
            for a in self.last_active:
                a.folder.widget.queue_draw()
        self.last_active = self.active
        #return True
    
    def preview_click_cb(self, widget, event):
        1
        #if event.button == 1:
            #if event.type == gtk.gdk._2BUTTON_PRESS:
                #self.execute_cb(widget, self.viewers[self.viewer_default])
    
    ####################
    # OTEHER CALLBACKS #
    ####################
    
    def counter_cb(self):
        self.clock.acquire()
        self.loaded += 1
        ld = self.loaded
        ql = self.queuelen
        if not self.reading_dir:
            if not self.progress.get_visible():
                self.progress.show()
            if ql:
                self.progress.set_fraction(float(ld) / ql)
            self.progress.set_text(self.lang.thumbnails_progress % (ld, ql))
            if ld >= ql:
                self.progress.hide()
                self.queuelen = self.loaded = 0
        self.clock.release()
        self.force()
        
    
    
    #################
    # PREVIEW STUFF #
    #################
    
    def request_preview(self, img):
        w_ = self.preview.allocation.width  - 2 * int(self.config.get("glimpse", "preview_border")) # widget width
        h_ = self.preview.allocation.height - 2 * int(self.config.get("glimpse", "preview_border")) # widget height
        _w = img.resolution[0]
        _h = img.resolution[1]
        if self.previewgen:
            self.previewgen.stop()
        self.previewgen = Preview(img, _w, _h, w_, h_, self.preview_sink)
    
    def preview_sink(self, img):
        self.set_preview(img)
    
    def set_preview(self, image):
        self.plock.acquire()
        try:
            self.previewimg = image
        finally:
            self.plock.release()
        self.preview.queue_draw()
        self.previewgen = None
    
    def expose_preview(self, widget, data):
        ctx = widget.window.cairo_create()
        try:
            w = widget.allocation.width - 2 * int(self.config.get("glimpse", "preview_border")) # widget width
            h = widget.allocation.height - 2 * int(self.config.get("glimpse", "preview_border")) # widget height
            ctx.set_source_rgba(*Styles[self.style]["image_background"])
            ctx.paint()
        except:
            return
        self.plock.acquire()
        try:
            if not self.previewimg:
                return
            x_   = int((w - self.previewimg.get_width()) / 2) + int(self.config.get("glimpse", "preview_border")) # x position of the final thumbnail on the folder surface
            y_   = int((h - self.previewimg.get_height()) / 2) + int(self.config.get("glimpse", "preview_border")) # y position of the final thumbnail on the folder surface
            
            ctx.set_source_pixbuf(self.previewimg, x_, y_)
            ctx.paint()
        finally:
            self.plock.release()
        
    
    ################
    # FOLDER STUFF #
    ################
    
    def redraw(self, folder, x = 0, y = 0, width = 0, height = 0):
        log("redraw (glimpse) %s %d %d %d %d" % (folder.name, x, y, width, height), 2)
        # request a redraw for a specified folder
        # state x, y, width and height to redraw just a portion of the
        # canvas
        if width and height:
            gobject.idle_add(folder.widget.queue_draw_area, x, y, width, height)
        else:
            gobject.idle_add(folder.widget.queue_draw)
        gobject.idle_add(self.force)
    
    def destroy_folder(self, folder):
        if folder is None:
            return
        log("destroy folder %s" % folder.path, 2)
        self.lock.acquire()
        try:
            self.folders.pop(folder.id)
            folder.destroy()
        finally:
            self.lock.release()
        self.rearrange_folders()
    
    def draw_folders(self):
        log("draw folders", 2)
        """trigger a draw on all folders at once"""
        self.lock.acquire()
        try:
            for f in self.folders:
                f.set_dirty(True)
        finally:
            self.lock.release()
    
    def rearrange_folders(self):
        log("rearange folders", 2)
        self.lock.acquire()
        try:
            self.folders = sorted(self.folders, key=lambda x: x.path.lower())
            _f = 0
            for f in self.folders:
                f.id = _f
                if f.widget.get_parent():
                    self.drawings.reorder_child(f.widget, _f)
                _f += 1
        finally:
            self.lock.release()
        
    
    def append_folder(self, path, fid = -1):
        log("append_folder %s" % path, 2)
        self.lock.acquire()
        try:
            if fid < 0:
                fid = len(self.folders)
            f = Folder(path, fid, self, self.drawing_click_cb, self.mousemove_cb, self.style, self.sizes[self.size])
            #if not len(f.images):
                #return None
            self.folders.append(f)
            self.drawings.pack_start(f.widget)
            self.queuelen += len(f.images)
        finally:
            self.lock.release()
        
        self.wlock.acquire()
        try:
            self.watch = dict(self.watch, **self.watchmanager.add_watch(path, self.watchmask, rec=False))
        finally:
            self.wlock.release()
            
        self.progress.set_text(self.lang.loading_progress % (self.queuelen, len(self.folders)))
        self.progress.pulse()
        self.force()
        return f
    
    def read_dir(self, directory, fid, single = False):
        log("read_dir %s" % directory, 2)
        # reads a directory recursively
        if fid == -1:
            fid = len(self.folders)
        
        _f = self.append_folder(directory, fid)
        fid += 1
        
        if not single:
            for __f in _f.folders:
                fid = self.read_dir(__f, fid, single)
        
        if fid and fid < len(self.folders) and self.folders[fid]["path"] == self.folders[fid - 1]["path"]:
            # if we re-read a folder (so the following one is the
            # same as this one but older) we have to destroy the
            # follow-up
            self.destroy_folder(self.folders[fid])
                
        return fid
        
    ###################
    # ACTIVE HANDLING #
    ###################
        
    def active_remove(self, image):
        self._active.remove(image.path)
        self.active.remove(image)
    
    def active_append(self, image):
        self._active.append(image.path)
        self.active.append(image)
        
    def active_set(self, lst):
        self.active = lst
        self._active = []
        for a in self.active:
            self._active.append(a.path)
        
    def get_active(self, path):
        # this is the joker why we do all this function fuckup around
        # active list - to have a fast way to check for actives in
        # read_dir()
        return path in self._active
        
    def find_active(self, path):
        # return a selected image or none for a given path
        if not self.get_active(path):
            return None
        for i in self.active:
            if i.path == path:
                return i
        return None
        
    #####################
    # HELPERS AND STUFF #
    #####################
    
    def force(self):
        while gtk.events_pending():
            gtk.main_iteration()
    
    def has_folder(self, path):
        # return a folder or none for a given path
        self.lock.acquire()
        try:
            for f in self.folders:
                if f.path == path:
                    return f
        finally:
            self.lock.release()
        return None
        
    def has_image(self, folder, image):
        # return an image or none for a path in a specified folder
        self.lock.acquire()
        try:
            folder.lock.acquire()
            try:
                for i in folder.images:
                    if i.path == image:
                        return i
            finally:
                folder.lock.release()
        finally:
            self.lock.release()
        return None
        
    def build_info(self, img, show_number = True, number = 0):
        # build the info text for an image
        txt = ""
        c = 0
        if img:
            img.folder.lock.acquire()
            try:
                for i in self.infos:
                    if c:
                        txt += (self.lang.infos[i] % getattr(img, i)) + "\n"
                    elif show_number and not c:
                        txt += (self.lang.infos[i] % (number)) + "\n"
                    c += 1
            finally:
                img.folder.lock.release()
        return txt
    
    def build_exif(self, img):
        # build the exif information for an image
        txt = ""
        l = 0
        if img and img.metadata:
            img.folder.lock.acquire()
            try:
                for i in img.metadata.exif_keys:
                    if not img.metadata[i].value:
                        continue
                    l = max(len(i), l)
                l += 2
                for i in img.metadata.exif_keys:
                    val = img.metadata[i].human_value.strip()
                    if not val:
                        continue
                    s = i + ":"
                    while len(s) < l:
                        s += " "
                    txt += s + val + "\n"
            finally:
                img.folder.lock.release()
        return txt
        
    def style_widget(self, widget):
        # copy a gtk style, set all background and foreground colors to
        # the ones according to the actually selected theme and apply
        # the new gtk style to the widget
        s = Styles[self.style]
        rb = int(s["folder_background"][0] * 65535)
        gb = int(s["folder_background"][1] * 65535)
        bb = int(s["folder_background"][2] * 65535)
        bg = (rb, gb, bb)
        rf = int(s["line_color"][0] * 65535)
        gf = int(s["line_color"][1] * 65535)
        bf = int(s["line_color"][2] * 65535)
        fg = (rf, gf, bf)
        
        st = widget.get_style().copy()
        
        st.bg[gtk.STATE_NORMAL]     = widget.get_colormap().alloc(*bg)
        st.bg[gtk.STATE_PRELIGHT]   = widget.get_colormap().alloc(*bg)
        st.bg[gtk.STATE_ACTIVE]     = widget.get_colormap().alloc(*bg)
        
        st.base[gtk.STATE_NORMAL]   = widget.get_colormap().alloc(*bg)
        st.base[gtk.STATE_PRELIGHT] = widget.get_colormap().alloc(*bg)
        st.base[gtk.STATE_ACTIVE]   = widget.get_colormap().alloc(*bg)
        
        st.fg[gtk.STATE_NORMAL]     = widget.get_colormap().alloc(*fg)
        st.fg[gtk.STATE_PRELIGHT]   = widget.get_colormap().alloc(*fg)
        st.fg[gtk.STATE_ACTIVE]     = widget.get_colormap().alloc(*fg)
        
        st.text[gtk.STATE_NORMAL]   = widget.get_colormap().alloc(*fg)
        st.text[gtk.STATE_PRELIGHT] = widget.get_colormap().alloc(*fg)
        st.text[gtk.STATE_ACTIVE]   = widget.get_colormap().alloc(*fg)
        
        widget.set_style(st)
    
    def key_event_to_string(self, event):
        name = ""
        if event.state & gtk.gdk.CONTROL_MASK:
            name += "CTRL+"
        if event.state & gtk.gdk.MOD1_MASK:
            name += "ALT+"
        if event.state & gtk.gdk.SHIFT_MASK:
            name += "SHIFT+"
        name += gtk.gdk.keyval_name(event.keyval)
        return name.upper()
    
    def key_event_to_unicode(self, event):
        name = ""
        if event.state & gtk.gdk.CONTROL_MASK:
            name += "CTRL+"
        if event.state & gtk.gdk.MOD1_MASK:
            name += "ALT+"
        if event.state & gtk.gdk.SHIFT_MASK:
            name += "SHIFT+"
        name += str(gtk.gdk.keyval_to_unicode(event.keyval))
        return name.upper()
        
    def write_config(self):
        f = open(self.configfile, "w")
        self.config.write(f);
        f.close()
    
    def start_config(self, cfgfile, defaults={}, sections=[],
                           optionxform_off = False):
        """Init a config objfrom commands import getoutputect
        and create config file if not existent"""
        if not os.path.exists(cfgfile):
            c = open(cfgfile, "w")
            c.close()
        cfg = config.RawConfigParser(defaults)
        if optionxform_off:
            cfg.optionxform = str
        cfg.readfp(open(cfgfile))
        for s in sections:
            if not cfg.has_section(s):
                cfg.add_section(s)
                for i in defaults:
                    cfg.set(s, i, defaults[i])
            else:
                for i in defaults:
                    try:
                        cfg.get(s, i)
                    except:
                        cfg.set(s, i, defaults[i])
        return cfg
        
    def set_language(self):
        st = "languages.%s" % self.language
        tmp = __import__(st, globals(), locals(), ['dictionary'], -1)
        self.lang = tmp.dictionary()
        
    def fill_language(self):
        # filechooser button
        #self.filechooser.set_title(self.lang.filechooser_folder)
        
        # style combobox
        s_ = self.style
        self.stylebox.get_model().clear()
        for s in Styles:
            s = Styles[s]
            try:
                n = self.lang.styles[s["name"]]
            except:
                n = s["name"]
            self.stylebox.append_text(n)
        self.stylebox.set_active(Styles.keys().index(s_))
        
        # size combobox
        s_ = self.size
        self.sizebox.get_model().clear()
        for s in self.sizes:
            try:
                n = self.lang.sizes[s["name"]]
            except:
                n = s["name"]
            self.sizebox.append_text(n)
        self.sizebox.set_active(s_)
        
        # toolbar
        self.sidebarbutton.set_label(self.lang.toolbar_sidebar)
        self.infobutton.set_label(self.lang.toolbar_info)
        
    
    
    def get_folder_by_path(self, path):
        f_ = None
        self.lock.acquire()
        try:
            for f in self.folders:
                if f.path == path:
                    f_ = f
                    break
        finally:
            self.lock.release()
        return f_
    
    def get_intid(self, img):
        return img.folder.id * self.up + img.id
    
    def get_folder_int(self, intid):
        return int(intid / self.up)
    
    def get_image_int(self, intid):
        return int(intid % self.up)
    
    def fire_mousemove(self):
        if self.lastmove:
            self.mousemove_cb(*self.lastmove) 
            
    def get_image(self, x, y, folder):
        s = Styles[self.style]
        folder.lock.acquire()
        try:
            for i in folder.images:
                try:
                    if (x > i.x and x < i.x + i.width - s["margin_col"]
                    and y > i.y and y < i.y + i.height - s["margin_row"]):
                        return i
                except:
                    continue
        finally:
            folder.lock.release()
        return None
        
    ##############
    # MENU STUFF #
    ##############
    
    def show_image_menu(self, x, y, item):
        # all menus and submenus
        self.menu = gtk.Menu()
        editmenu = gtk.Menu()
        viewmenu = gtk.Menu()
        
        # submenu edit/view buttons
        editsub = gtk.MenuItem(self.lang.editor_submenu)
        viewsub = gtk.MenuItem(self.lang.viewer_submenu)
        
        editsub.set_submenu(editmenu)
        viewsub.set_submenu(viewmenu)
        
        # edit submenu items
        hasedit = False
        for e in self.editors:
            if e[0] != self.editors[self.editor_default][0] and self.which(e[1]):
                mi = gtk.MenuItem(e[0])
                mi.connect("activate", self.execute_cb, e)
                editmenu.append(mi)
                hasedit = True
        # view submenu items
        hasview = False
        for v in self.viewers:
            if v[0] != self.viewers[self.viewer_default][0] and self.which(v[1]):
                mi = gtk.MenuItem(v[0])
                viewmenu.append(mi)
                mi.connect("activate", self.execute_cb, v)
                hasview = True
        
        # folder
        folder = gtk.MenuItem(self.lang.folder_open)
        folder.connect("activate", self.folder_cb)
        
        # mailer
        hasmail = False
        for m in self.mailers:
            if self.which(m[1]):
                mailto = gtk.MenuItem(self.lang.send_to % m[0])
                mailto.connect("activate", self.send_to_cb, m)
                hasmail = True
        
        # default edit button
        edit = gtk.MenuItem(self.lang.editor_default % self.editors[self.editor_default][0])
        edit.connect("activate", self.execute_cb, self.editors[self.editor_default])
        
        #default view button
        view = gtk.MenuItem(self.lang.viewer_default % self.viewers[self.viewer_default][0])
        view.connect("activate", self.execute_cb, self.viewers[self.viewer_default])
        
        # rename
        rename = gtk.MenuItem(self.lang.rename_menu)
        rename.connect("activate", self.rename_cb)
        
        # delete
        delete = gtk.MenuItem(self.lang.delete_menu)
        delete.connect("activate", self.delete_cb)
        
        # copy
        copy = gtk.MenuItem(self.lang.copy_menu)
        copy.connect("activate", self.copy_cb)
        
        # move
        move = gtk.MenuItem(self.lang.move_menu)
        move.connect("activate", self.move_cb)
        
        # stitch together
        
        #############################
        # DEFAULT VIEW
        if self.which(self.viewers[self.viewer_default][1]):
            self.menu.append(view)
        # SUBMENU VIEW
        if hasview:
            self.menu.append(viewsub)
        ############################
        self.menu.append(gtk.SeparatorMenuItem())
        # DEFAULT EDIT
        if self.which(self.editors[self.editor_default][1]):
            self.menu.append(edit)
        # SUBMENU EDIT
        if hasedit:
            self.menu.append(editsub)
        #############################
        self.menu.append(gtk.SeparatorMenuItem())
        self.menu.append(folder)
        
        if hasmail:
             self.menu.append(mailto)
        
        #############################
        self.menu.append(gtk.SeparatorMenuItem())
        # RENAME
        self.menu.append(rename)
        # DELETE
        self.menu.append(delete)
        # COPY
        self.menu.append(copy)
        # MOVE
        self.menu.append(move)
        #############################
        
        self.menu.show_all()
        self.menu.popup(None, None, None, 3, gtk.get_current_event_time())
    
    
    ################
    # WINDOW STUFF #
    ################
    def ctrl_c(self, signum, frame):
        self.delete_event(None, None)
        self.destroy(None)
    
    def delete_event(self, widget, event, data=None):
        self.destroying = True
        self.notifier.stop()
        self.config.set("glimpse", "win_x", self.window.get_position()[0])
        self.config.set("glimpse", "win_y", self.window.get_position()[1])
        self.config.set("glimpse", "win_w", self.window.get_size()[0])
        self.config.set("glimpse", "win_h", self.window.get_size()[1])
        self.config.set("glimpse", "hpaned", self.toppaned.get_position() if self.toppaned.get_position() else self.bottompaned.get_position())
        self.config.set("glimpse", "vpaned", self.mainpaned.get_position())
        self.write_config()
        return False
        
    def destroy(self, widget, data=None):
        if self.window:
            self.window.destroy()
        if gtk.main_level():
            gtk.main_quit()
        exit()
   
    def _response(self, entry, dialog, response):
        dialog.response(response)
    
    def dialog(self, title = "", description = ""):
        dialog = gtk.MessageDialog(
            None,
            gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
            gtk.MESSAGE_QUESTION,
            gtk.BUTTONS_OK_CANCEL,
            None)
        dialog.set_title(title)
        dialog.set_markup(description)
        dialog.set_default_response(gtk.RESPONSE_OK)
        response = dialog.run()
        if response == gtk.RESPONSE_CANCEL:
            dialog.destroy()
            return False
        dialog.destroy()
        return True
    
    def _entry_preview(self, widget, data):
        data[1].set_text(data[0](widget.get_text()))
        
    def entry_dialog(self, title = "", name = "", description = "", preset = "", select = (-1, -1), hlp = "", htitle = "", preview = None):
        #base this on a message dialog
        dialog = gtk.MessageDialog(
            None,
            gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
            gtk.MESSAGE_QUESTION,
            gtk.BUTTONS_OK_CANCEL,
            None)
        dialog.set_title(title)
        dialog.set_markup(description)
        dialog.set_default_response(gtk.RESPONSE_OK)
        
        #create the text input field
        entry = gtk.Entry()
        entry.set_text(preset)
             
        #allow the user to press enter to do ok
        entry.connect("activate", self._response, dialog, gtk.RESPONSE_OK)
        
        #create a horizontal box to pack the entry and a label
        hbox = gtk.HBox()
        hbox.pack_start(gtk.Label(name), False, 5, 5)
        hbox.pack_end(entry)
        dialog.vbox.pack_start(hbox, True, True, 0)
        
        # preview label
        if preview:
            pl = gtk.Label()
            dialog.vbox.pack_start(pl)
            entry.connect("changed", self._entry_preview, (preview, pl))
            pl.set_text(preview(preset))
        
        # help?
        if hlp and htitle:
            ht = gtk.Label()
            ht.set_markup(hlp)
            ex = gtk.Expander(htitle)
            ex.set_expanded(False)
            ex.add(ht)
            dialog.vbox.pack_start(ex, True, True, 0)
            
        dialog.show_all()
        
        if select[0] + select[1] >= 0:
             entry.select_region(*select)
             
        #go go go
        response = dialog.run()
        if response == gtk.RESPONSE_CANCEL:
            dialog.destroy()
            return
        text = entry.get_text()
        dialog.destroy()
        return text
        
    def overwrite_dialog(self, title="", text="", name="", preset="", select=(-1,-1), multi=False):
        if multi:
            buttons = (self.lang.overwrite_all, gtk.RESPONSE_ACCEPT,
                       self.lang.overwrite_cancel_all, gtk.RESPONSE_REJECT)
        else:
            buttons = ()
        buttons += (self.lang.overwrite_cancel, gtk.RESPONSE_CANCEL,
                    self.lang.overwrite_ok, gtk.RESPONSE_OK)
        dialog = gtk.MessageDialog(
            self.window,
            gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
            gtk.MESSAGE_WARNING,
            0)
        dialog.set_title(title)
        dialog.add_buttons(*buttons)
        
        dialog.set_markup(text)
        
        #create the text input field
        entry = gtk.Entry()
        entry.set_text(preset)
            
        #allow the user to press enter to do ok
        entry.connect("activate", self._response, dialog, gtk.RESPONSE_OK)
        
        #create a horizontal box to pack the entry and a label
        hbox = gtk.HBox()
        hbox.pack_start(gtk.Label(name), False, 5, 5)
        hbox.pack_end(entry)
        dialog.vbox.pack_start(hbox, True, True, 0)
        
        dialog.show_all()
        
        response = dialog.run()
        
        r = 0
        if response == gtk.RESPONSE_OK:
            # rename/overwrite single
            r = entry.get_text()
        if response == gtk.RESPONSE_REJECT:
            # cancel all
            r = -1
        if response == gtk.RESPONSE_CANCEL:
            # cancel single
            r = 0
        if response == gtk.RESPONSE_ACCEPT:
            # overwrite all
            r = 1
        dialog.destroy()
        return r
        
    def help(self):
        win = gtk.Window()
        win.set_title(self.lang.help_title)
        win.set_icon_from_file(os.path.join(self.progdir, "glimpse.png"))
        win.connect("destroy", self.help_cb)
        sc = gtk.ScrolledWindow()
        sc.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_ALWAYS)
        ta = gtk.TextView()
        ta.set_editable(False)
        ta.modify_font(pango.FontDescription('monospace'))
        ta.get_buffer().set_text(self.lang.help)
        ta.set_border_width(10)
        win.add(sc)
        sc.add(ta)
        win.resize(620, 400)
        win.show_all()
        self.style_widget(ta)
        return win
        
    ##################
    # OPTIONS WINDOW #
    ##################
    
    def options_window(self):
        if not self.options:
            self.options = gtk.Window(gtk.WINDOW_TOPLEVEL)
            self.options.connect("destroy", self.ow_destroy)
            self.options.set_title(self.lang.options_title)
            self.options.set_icon_from_file(os.path.join(self.progdir, "glimpse.png"))
            
            t = gtk.Table(7, 2)
            t.set_border_width(10)
            t.set_col_spacings(5)
            t.set_row_spacings(10)
            
            # labels
            l = gtk.Label(self.lang.options_language)
            l.set_alignment(1, 0.5)
            t.attach(l, 0, 1, 0, 1)
            l = gtk.Label(self.lang.options_default_view)
            l.set_alignment(1, 0.5)
            t.attach(l, 0, 1, 1, 2)
            l = gtk.Label(self.lang.options_default_edit)
            l.set_alignment(1, 0.5)
            t.attach(l, 0, 1, 2, 3)
            l = gtk.Label(self.lang.options_autoclean)
            l.set_alignment(1, 0.5)
            t.attach(l, 0, 1, 3, 4)
            l = gtk.Label(self.lang.options_preview)
            l.set_alignment(1, 0.5)
            t.attach(l, 0, 1, 4, 5)
            l = gtk.Label(self.lang.options_cache)
            l.set_alignment(1, 0.5)
            t.attach(l, 0, 1, 5, 6)
            
            # language
            c = gtk.combo_box_new_text()
            act = i = 0
            for loc in sorted(self.languages.iterkeys()):
                c.append_text("%s (%s)" % (self.languages[loc], loc))
                if loc == self.language:
                    act = i
                i += 1
            c.set_active(act)
            t.attach(c, 1, 2, 0, 1)
            c.connect('changed', self.language_cb)
            
            # default view
            c = gtk.combo_box_new_text()
            c.connect('changed', self.defview_cb)
            for v in self.viewers:
                c.append_text(v[0])
            c.set_active(self.viewer_default)
            t.attach(c, 1, 2, 1, 2)
            
            # default edit
            c = gtk.combo_box_new_text()
            c.connect('changed', self.defedit_cb)
            for v in self.editors:
                c.append_text(v[0])
            c.set_active(self.editor_default)
            t.attach(c, 1, 2, 2, 3)
            
            # autoclean
            s = gtk.SpinButton()
            s.set_digits(0)
            s.set_increments(1, 10)
            s.set_range(0, 10000000)
            s.set_value(int(float(self.config.get("glimpse", "autoclean"))))
            s.connect("value-changed", self.autoclean_cb)
            t.attach(s, 1, 2, 3, 4)
            
            # preview
            s = gtk.CheckButton()
            s.set_active(int(self.config.get("glimpse", "preview")))
            s.connect("clicked", self.preview_cb)
            t.attach(s, 1, 2, 4, 5)
            
            # cache
            s = gtk.CheckButton()
            s.set_active(int(self.config.get("glimpse", "cache")))
            s.connect("clicked", self.cache_cb)
            t.attach(s, 1, 2, 5, 6)
            
            # close button
            b = gtk.Button()
            b.set_label(self.lang.options_close)
            b.connect("clicked", self.ow_destroy)
            t.attach(b, 0, 2, 7, 8)
            
            self.options.add(t)
            self.options.show_all()
        else:
            self.options.destroy()
            self.options = None
    
    def ow_destroy(self, window):
        self.write_config()
        if self.options:
            self.options.destroy()
        self.options = None
        
    def language_cb(self, widget):
        model = widget.get_model()
        lan = model.get_value(widget.get_active_iter(), 0)
        match = re.search(r"\(([a-zA-Z_]+)\)$", lan.strip())
        self.language = match.group(1)
        self.config.set("glimpse", "language", self.language)
        self.set_language()
        self.fill_language()
    
    def defview_cb(self, widget):
        self.viewer_default = widget.get_active()
        self.config.set("glimpse", "viewer_default", self.viewer_default)
        
    def defedit_cb(self, widget):
        self.editor_default = widget.get_active()
        self.config.set("glimpse", "editor_default", self.editor_default)
    
    def autoclean_cb(self, widget):
        self.config.set("glimpse", "autoclean", widget.get_value_as_int())
    
    def sidebar_cb(self, widget):
        self.config.set("glimpse", "sidebar", int(widget.get_active()))
        self.set_sidebar(widget.get_active())
    
    def infobar_cb(self, widget):
        self.config.set("glimpse", "infobar", int(widget.get_active()))
        self.set_infobar(widget.get_active())
    
    def preview_cb(self, widget):
        self.config.set("glimpse", "preview", int(widget.get_active()))
        self.show_preview(widget.get_active())
    
    def cache_cb(self, widget):
        self.config.set("glimpse", "cache", int(widget.get_active()))
        
        
if __name__ == "__main__":
    glimpse = Glimpse()
    gtk.main()
