#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) Markus Schmidt <schmidt@boomshop.net>
# License: F4A (free4all) - have a look in the LICENSE file that came with this
# software

import os
import re
import sys
import gtk
import math
import cairo
import mimetypes
import threading

from lib.Utils import Utils
from lib.Image import Image
from lib.Styles import Styles
from lib.Utils import Utils

log_level = 1
def log(mesg, level = 0):
    if level <= log_level:
        print mesg
        
class Folder(Utils):
    
    id      = -1
    name    = ""
    path    = ""
    parent  = ""
    images  = []
    folders = []
    dirty   = False
    resized = False
    width   = -1
    height  = -1
    size    = None
    rows    = 0
    cols    = 0
    widget  = None
    glimpse = None
    style   = None
    widget  = None
    buffer  = None
    bufc    = None
    thumb_width  = 0
    thumb_height = 0
    thumbnailer  = None
    
    lock = threading.Lock()
    
    types = ["bmp", "ico", "jpeg", "png", "tga", "tiff"] # all readable image types
    
    def __init__(self, directory, id, glimpse, click = None, move = None, style = None, size = None):
        log("__init__ FOLDER %s" % directory, 2)
        self.glimpse = glimpse
        self.thumbnailer = glimpse.thumbnailer
        self.id = id
        self.widget  = gtk.DrawingArea()
        self.widget.add_events(gtk.gdk.BUTTON_PRESS_MASK
                             | gtk.gdk.BUTTON_RELEASE_MASK
                             | gtk.gdk.POINTER_MOTION_MASK
                             | gtk.gdk.POINTER_MOTION_HINT_MASK )
        
        self.widget.connect("expose-event", self.expose)
        self.glimpse.style_widget(self.widget)
        
        if move:
            self.widget.connect("motion_notify_event", move, self)
        if click:
            self.widget.connect("button_press_event", click, self)
                             
        self.widget.show()
        
        if style:
            self.set_style(style, True)
            
        if size is not None:
            self.set_size(size, True)
        
        self._set_thumbnail_sizes()
        self.set_root(directory)
        
    ###########
    # BACKEND #
    ###########
    
    def set_root(self, folder):
        log("set_root (folder) %s" % folder, 2)
        """set the folders root dir and trigger everything neccessary
           to read and show the contents"""
        self.path   = folder
        self.name   = os.path.basename(folder)
        self.parent = os.path.dirname(folder)
        self.lock.acquire()
        try:
            self.images = []
        finally:
            self.lock.release()
        self.widget.set_name(self.name)
        #if not
        self.read()#:
            #self.destroy()
            #return False
        return True
        
    def read(self):
        log("read (folder) %s" % self.name, 2)
        """read the root directory and create Image instances for all
           images found"""
        self.dirty = True
        l = os.listdir(self.path)
        l.sort(key = str.lower)
        c = 0
        self.folders = []
        for f in l:
            # read all files
            path = os.path.join(self.path, f)
            # continue if it's a directory
            if os.path.isdir(path):
                self.folders.append(path)
                continue
            # ask which kind of file we have here and skip this cycle if
            # it's not a usable image
            type__ = mimetypes.guess_type(path)[0]
            if type__:
                type_ = [type__.split("/")[0], type__.split("/")[1].lower()]
            else:
                type_ = ["unknown", "???"]
            if type_[0] != "image" or not type_[1] in self.types:
                continue
            # seems to be a read- and usable image file
            # so create a new Image instance
            self.lock.acquire()
            try:
                self.images.append(Image(path, self.count(), self, type_))
            finally:
                self.lock.release()
            c += 1
        self.folders.sort(key=str.lower)
        #if not c:
            #return False
        self.resize()
        return True
    
    ############
    # EXPOSING #
    ############
    
    def draw_folder(self):
        log("draw_folder (folder) %s" % self.name, 2)
        """draw a folder without its contents - just the headlines and
           backgrounds"""
        s = Styles[self.style]
        if not self.bufc:
            return
        cr = self.bufc
        
        cr.set_source_rgba(*s["folder_background"])
        cr.rectangle(0, 0, self.width, self.height)
        cr.fill()
        
        if len(self.images):
            cr.set_source_rgba(*s["header_color"])
        else:
            cr.set_source_rgba(*s["header_empty"])
        cr.select_font_face(*s["header_font"])
        cr.set_font_size(s["header_size"])
        cr.move_to(s["margin_out"], s["header_height"] * 0.6)
        n = (self.glimpse.lang.header if len(self.images) else self.glimpse.lang.header_empty) % self.name
        cr.show_text(n)
        
        cr.set_font_size(s["line_size"])
        cr.set_source_rgba(*s["line_color"])
        cr.move_to(s["margin_out"], s["header_height"] * 0.95)
        cr.show_text(self.path)
    
    def draw_image(self, image):
        log("draw_image (folder) %s" % image.name, 2)
        """draw an image onto the surface"""
        img  = image.image
        cr   = self.bufc
        pre  = self.glimpse.sizes[len(self.glimpse.sizes) - 1]["size"]
        s    = Styles[self.style]
        size = self.size["size"]
        show = self.size["show"]
        
        if not img:
            img = self.size["image"]
        if type(img) is str:
            img = cairo.ImageSurface.create_from_png(img)
        
        if image.image:
            __w = image.resolution[0] # original image width
            __h = image.resolution[1] # original image height
        else:
            __w = pre # loading thumb width
            __h = pre # loading thumb height
            
        _w   = img.get_width()  # width of the saved thumbnail or the loading image
        _h   = img.get_height() # height of the saved thumbnail or the loading image
        
        rat  = float(_w) / _h # ratio of the thumbnail image
        rx   = 1 if rat > 1 else float(_w) / _h
        ry   = 1 if rat < 1 else float(_h) / _w
        w_   = min(_w, int(size * rx)) # (max) width of the final thumbnail
        h_   = min(_h, int(size * ry)) # (max) height of the final thumbnail
        x_   = image.x + (size - w_) / 2 + s["margin_img"] # x position of the final thumbnail on the folder surface
        y_   = image.y + (size - h_) / 2 + s["margin_img"] # y position of the final thumbnail on the folder surface
        
        cr.rectangle(image.x, image.y, image.width, image.height)
        cr.clip()
        
        # folder background
        cr.set_source_rgba(*s["folder_background"])
        cr.rectangle(image.x, image.y, image.width, image.height)
        cr.fill()
        
        # image background
        cr.set_source_rgba(*s["image_background"])
        cr.rectangle(image.x + s["margin_img"], image.y + s["margin_img"], size, size)
        cr.fill()
        
        # thumbnail
        tg = cairo.ImageSurface(cairo.FORMAT_ARGB32, w_, h_)
        ctx = cairo.Context(tg)
        ctx.scale(min(1.0, float(w_) / float(_w)), min(1.0, float(h_) / float(_h)))
        ctx.set_source_surface(img, 0, 0)
        ctx.set_operator(cairo.OPERATOR_SOURCE)
        ctx.paint()
        
        cr.set_source_surface(tg, x_, y_)
        cr.rectangle(image.x, image.y, image.width, image.height)
        cr.fill()
        
        # rotation icon
        if image.rotation > 1:
            cr.set_source_surface(self.glimpse.orient_icons[image.rotation - 2], image.x + image.width - s["icon_x"], image.y + s["icon_y"])
            cr.rectangle(image.x, image.y, image.width, image.height)
            cr.fill()
            
        # informations
        top = image.y + size + s["margin_img"] * 2
        ff  = float(self.glimpse.config.get("glimpse", "font_factor"))
        fs  = s["line_size"] * ff + ((s["line_size"] * (1 - ff)) * (float((self.glimpse.size + 1)) / float(len(self.glimpse.sizes))))
        lh  = s["line_height"] * ff + ((s["line_height"] * (1 - ff)) * (float((self.glimpse.size + 1)) / float(len(self.glimpse.sizes))))
        
        d = 0
        cr.set_font_size(fs)
        cr.set_source_rgba(*s["line_color"])
        for l in show:
            if l == show[0]:
                cr.select_font_face(*s["title_font"])
                cr.set_source_rgba(*s["title_color"])
            else:
                cr.select_font_face(*s["line_font"])
                cr.set_source_rgba(*s["line_color"])
            cr.move_to(image.x + s["margin_img"], top + d * lh + lh * 0.8)
            cr.show_text(self.shorten(str(getattr(image, l)), cr, size))
            d += 1
        
        image.dirty = False
        cr.reset_clip()
        
    def expose(self, widget, *data):
        log("expose (folder) %s %s" % (self.name, str(data)), 2)
        """expose callback for gtk
        redraws folders if they're marked as dirty and copies its
        cache drawingarea onto the final window"""
        s = Styles[self.style]
        ctx = widget.window.cairo_create()
            
        # folder marked as resized? prepare stuff
        #if self.resized:
            #self.prepare()
        
        # folder marked as dirty? redraw completely
        if self.dirty:
            self.draw_folder()
        
        # draw dirty images (or everything if folder is dirty)
        for i in self.images:
            if i.dirty or self.dirty:
                self.lock.acquire()
                try:
                    self.draw_image(i)
                    i.dirty = False
                finally:
                    self.lock.release()
        
        # copy surface to DrawingArea window
        ctx.set_source_surface(self.buffer, 0, 0)
        ctx.rectangle(0, 0, self.width, self.height)
        ctx.fill()
        
        # mark hovered
        if self.glimpse.hover and self.glimpse.hover.folder == self:
            i = self.glimpse.hover
            ctx.rectangle(i.x, i.y, i.width - s["margin_col"], i.height - s["margin_row"])
            ctx.set_source_rgba(*s["hover_color"])
            ctx.fill()
        
        # mark actives
        for a in self.glimpse.active:
            if a.folder == self:
                ctx.rectangle(a.x, a.y, a.width - s["margin_col"], a.height - s["margin_row"])
                ctx.set_source_rgba(*s["active_color"])
                ctx.set_line_width(s["active_border"])
                ctx.stroke_preserve()
                ctx.set_source_rgba(*s["active_blinder"])
                ctx.fill()
        
        #self.resized = False
        self.dirty = False
        return False
    
    ####################
    # STRUCTURAL STUFF #
    ####################
        
    def rearrange(self):
        log("rearrange (folder) %s" % self.name, 2)
        """resort all images in this folder"""
        self.images = sorted(self.images, key=lambda x: x.path.lower())
        c = 0
        self.lock.acquire()
        try:
            for i in self.images:
                i.set_id(c)
                c += 1
        finally:
            self.lock.release()
        self.dirty = True
        self.redraw()
    
    def redraw(self, image = None):
        log("redraw (folder) %s %s" % (self.name, image.name if image else ""), 2)
        """queue a redraw of the folders DrawingArea. if an image is set
           just redraw the portion of the surface where it is located"""
        if image:
            self.glimpse.redraw(self, image.x, image.y, image.width, image.height)
        else:
            self.glimpse.redraw(self)
    
    def prepare(self):
        log("prepare (folder) %s" % self.name, 2)
        """resize buffer and DrawingArea widget"""
        self.buffer = cairo.ImageSurface(cairo.FORMAT_ARGB32, self.width, self.height)
        self.bufc   = cairo.Context(self.buffer)
        self.widget.set_size_request(self.width, self.height)
        
    def refresh(self):
        log("refresh (folder) %s" % self.name, 2)
        """re-read the folder completely, drop everything else"""
        self.images = []
        self.dirty  = True
        self.read()
    
    def resize(self):
        log("resize (folder) %s" % self.name, 2)
        """should be called after the viewport resizes. calculates the
           dimensions of the surface and the amount of columns and rows
           showing the thumbnails."""
        self.lock.acquire()
        try:
            self._set_thumbnail_sizes()
            s           = Styles[self.style]
            self.width  = int(self.glimpse.viewport.allocation.width - self.glimpse.vmargin) # width of the folder surface
            w_          = self.width - s["margin_out"] * 2 # width without margins
            self.cols   = int((w_ + s["margin_col"]) / self.thumb_width) if len(self.images) else 0 # amount of thumbs per row
            self.rows   = int(math.ceil(float(len(self.images)) / float(self.cols))) if self.cols else 0 # amount of rows
            self.height = int(s["header_height"] + s["margin_out"] * 2 + self.thumb_height * self.rows - s["margin_row"]) # height of the canvas
            for i in self.images:
                i.set_allocation(*self.get_allocation(i.id))
        finally:
            self.lock.release()
        self.dirty   = True
        #self.resized = True
        self.prepare()
        self.redraw()
        
    #######################
    # GETTERS AND SETTERS #
    #######################
    
    def set_style(self, style, hold = False):
        """hand over a style dictionary to use with this folder"""
        self.style = style
        self.dirty = True
        if not hold:
            self.resize()
    
    def set_size(self, size, hold = False):
        """set a new size for the thumbnails drawn on this folders
           surface"""
        self.size  = size
        self.dirty = True
        if not hold:
            self.resize()
            
    def set_dirty(self, dirty, hold = False):
        """change the state of the dirty flag to eventually trigger a
           redraw of the folders surface and all of its thumbnails"""
        self.dirty = dirty
        if not hold:
            self.redraw()
    
    def count(self):
        """returns the amount of images contained into this folder"""
        return len(self.images)
        
    def get_allocation(self, id):
        """return an images allocation according to its position in the
           array. returns a tuple of (x, y, width, height, position
           horizonal, position vertical)"""
        s    = Styles[self.style]
        posh = id % self.cols if self.cols else 0; # numerical vertical position of image
        posv = int(id / self.cols if self.cols else 0); # numerical horizontal position of image
        x    = posh * (self.thumb_width) + s["margin_out"]; # x position of overall box on folder surface
        y    = posv * (self.thumb_height) + s["margin_out"] + s["header_height"]; # y position of overall box on folder surface
        return (x, y, self.thumb_width, self.thumb_height, posh, posv)
    
    def has_image(self, path):
        self.lock.acquire()
        try:
            for i in self.images:
                if i.path == path:
                    return i
        finally:
            self.lock.release()
        return False
    
    #############
    # INTERNALS #
    #############
    
    def _set_thumbnail_sizes(self):
        """is used internally after some values changed to lower the cpu
           usage when asking for allocations and stuff"""
        s = Styles[self.style]
        self.thumb_width  = self.size["size"] + 2 * s["margin_img"] + s["margin_col"] # thumbnail area width
        self.thumb_height = self.size["size"] + 3 * s["margin_img"] + s["margin_row"] + s["line_height"] * len(self.size["show"]) # thumbnail area height
    
    def destroy(self):
        log("destroy (folder) %s" % self.name, 2)
        """destroy a folder and all of its contents"""
        self.lock.acquire()
        try:
            for i in self.images:
                i.destroy();
        finally:
            self.lock.release()
        self.widget.destroy()
        self.images = []
        self.buffer = None
        self.bufc   = None
