# Copyright (C) Markus Schmidt <schmidt@boomshop.net>
# License: F4A (free4all) - have a look in the LICENSE file that came with this
# software

import os
import hashlib
import pyexiv2
import threading

from lib.Thumbnailer import Thumbnailer
from lib.Utils import Utils

log_level = 1
def log(mesg, level = 0):
    if level <= log_level:
        print mesg
        
class Image(Utils):
    id          = -1 # position in the folders array
    folder      = None # parent folder object
    name        = "" # name of the image
    path        = "" # path including name of the image
    parent      = "" # the folders path containing the image
    type        = "" # type of image, e.g. png, jpeg, ...
    md5         = "" # md5 string of the original image
    size        = 0 # file size in bytes of the image
    sizeh       = "0 Bytes" # human readable file size
    hash        = "" # md5 string of the path to generate thumbnail name 
    resolution  = (0, 0) # width and height of the original image
    resolutionh = "0x0" # resolution as human readable string
    image       = None # cairo context pointing to a surface with a ready-to-use thumbnail or None to use loading image
    rotation    = 1 # rotation of the image http://sylvana.net/jpegcrop/exif_orientation.html
    thumbnailer = None # thumbnailer instance which is used for generation
    x           = 0 # position of the thumbnail box
    y           = 0 # position of the thumbnail box
    width       = 0 # width of the thumbnail box
    height      = 0 # height of the thumbnail box
    posh        = 0 # position in the thumbnail grid in x direction
    posv        = 0 # position in the thumbnail grid in y direction
    metadata    = None # exif metadata object
    dirty       = False # trigger a complete redraw of the image
    lock        = threading.Lock()
    
    def __init__(self, path, id, folder, type_):
        log("__init__ IMAGE %s" % path, 2)
        self.folder = folder
        self.type = type_[1]
        self.set_id(id)
        self.set_path(path)
    
    def refresh(self):
        """trigger a complete redraw of the thumbnail"""
        self.thumbnailer = self.folder.thumbnailer
        self.thumbnailer.add(self)
        
    def set_path(self, path, force = False):
        """set the path of the image and trigger some neccessary stuff
           to show the new image"""
        # quit if nothing changed and we aren't forced
        if self.path == path and not force:
            return
        self.dirty  = True
        self.path   = path
        self.name   = os.path.basename(path)
        self.parent = os.path.dirname(path)
        self.hash   = str(int(hashlib.md5(self.path).hexdigest(), 16))
        self.refresh()
    
    def set_id(self, id):
        """set the position in the folders array"""
        self.id = id
        self.set_allocation(*self.folder.get_allocation(id))
        
    def set_allocation(self, x, y, width, height, posh, posv):
        """set the image boxes allocation"""
        self.x      = x
        self.y      = y
        self.width  = width
        self.height = height
        self.posh   = posh
        self.posv   = posv
    
    def destroy(self):
        log("destroy (image) %s" % self.name, 2)
        """destroy an image and stop an eventually running thumbnailer"""
        if self.thumbnailer:
            self.thumbnailer.stop()
        self.thumbnailer = None
        self.image       = None
