# Copyright (C) Markus Schmidt <schmidt@boomshop.net>
# License: F4A (free4all) - have a look in the LICENSE file that came with this
# software

import pyinotify
from lib.Thumbnail import Thumbnail

log_level = 1
def log(mesg, level = 0):
    if level <= log_level:
        print mesg
        
class Notifier(pyinotify.ProcessEvent):
    def __init__(self, parent):
        self.parent = parent
        

    def process_IN_MODIFY(self, event):
        log("Modifying: %s" % event.pathname, 2)
        if event.dir:
            # directory
            f = self.parent.has_folder(event.path)
            if f:
                f.set_dirty(True)
        else:
            #image file
            f = self.parent.has_folder(event.path)
            if f:
                f.rearrange()
    
    def process_IN_CREATE(self, event):
        log("Create: %s" % event.pathname, 2)
        self.create(event)
        
    def process_IN_MOVED_TO(self, event):
        log("Moved To: %s" % event.pathname, 2)
        self.create(event)
    
    def process_IN_MOVED_FROM(self, event):
        log("Moved From: %s" % event.pathname, 2)
        self.delete(event)
    
    def process_IN_DELETE(self, event):
        log("Delete: %s" % event.pathname, 2)
        self.delete(event)
    
    
    def create(self, event):
        if event.dir:
            # directory
            self.parent.read_dir(event.pathname, -1, False)
            self.parent.rearrange_folders()
        else:
            # image file
            f = self.parent.has_folder(event.path)
            if not f:
                # folder isn't in the list (e.g. cause it was empty)
                # so add the folder to the list
                self.parent.read_dir(event.pathname, -1, False)
                self.parent.rearrange_folders()
            else:
                f.refresh()
    
    
    def delete(self, event):
        if event.dir:
            # directory
            item = self.parent.has_folder(event.pathname)
            if item:
                self.parent.destroy_folder(item)
        else:
            # image file
            fold = self.parent.has_folder(event.path)
            if not fold:
                return
            item = self.parent.has_image(fold, event.pathname)
            if not item:
                return
            fold.lock.acquire()
            try:
                item = fold.images.pop(item.id)
            finally:
                fold.lock.release()
            fold.rearrange()
            fold.resize()
