# Copyright (C) Markus Schmidt <schmidt@boomshop.net>
# License: F4A (free4all) - have a look in the LICENSE file that came with this
# software

import os
import cairo
import hashlib
import gobject
import pyexiv2
from threading import Thread
from PIL import Image
from lib.Utils import Utils

log_level = 1
def log(mesg, level = 0):
    if level <= log_level:
        print mesg
        
gobject.threads_init()

class Thumbnail(Thread):
    stopped = False
    def __init__(self, queue, lock, thumbnailer, width, height, cb_ready, cb_fail, cb_done):
        Thread.__init__(self)
        log("__init__ THUMBNAIL %d %d" % (thumbnailer, len(queue)), 2)
        self.queue       = queue
        self.lock        = lock
        self.thumbnailer = thumbnailer
        self.ready       = cb_ready
        self.fail        = cb_fail
        self.done        = cb_done
        self.width       = width
        self.height      = height
        self.start()
    
    def stop(self):
        self.stopped = True
        
    def _return(self, image, targ, md5 = None, resolution = None):
        log("_return (thumbnail) %s" % image.name, 2)
        if self.stopped:
            self.fail(image)
            return
            
        if not md5 or not resolution:
            data = open(targ, "r").readlines()
        if not md5:
            md5 = data[0].rstrip()
        if not resolution:
            resolution = data[1].rstrip()
        image.resolution = (resolution.split("x")[0], resolution.split("x")[1])
        image.resolutionh = resolution
        image.md5 = md5
        if not int(image.folder.glimpse.config.get("glimpse", "cache")):
            image.image = targ + ".png"
        else:
            image.image = cairo.ImageSurface.create_from_png(targ + ".png")
        if image.metadata and "Exif.Image.Orientation" in image.metadata.exif_keys:
            image.rotation = int(image.metadata["Exif.Image.Orientation"].value)
        image.dirty = True
        self.ready(image)
    
    def _read_metadata(self, image):
        """used internally to read and store all interesting data around
           the image like size and exif meta data"""
        image.size   = os.stat(image.path).st_size
        image.sizeh  = image.bytes2string(image.size) + "Bytes"
        try:
            # try to read exif metadata
            metadata = pyexiv2.ImageMetadata(image.path)
            metadata.read()
            image.metadata = metadata
        except:
            # or set the member to None
            image.metadata = None
            
    def run(self):
        log("run (thumbnail) %d" % len(self.queue), 2)
        while 1:
            self.lock.acquire()
            try:
                if not len(self.queue):
                    self.done(self.thumbnailer)
                    break
                image = self.queue.pop(0)
                targ  = os.path.join(image.folder.glimpse.thumbdir, image.hash)
            finally:
                self.lock.release()
            
            if not os.path.isfile(image.path):
                # no source image available
                self.fail(image)
                continue
                
            self._read_metadata(image)
            
            if (not os.path.isfile(targ + ".png")
            or  not os.path.isfile(targ)
            or not  os.path.getsize(targ + ".png") > 100):
                # thumbnail doesn't exist
                self.thumbnail(image, targ)
                continue
            
            if os.path.getmtime(targ) <= os.path.getmtime(image.path):
                # thumbnail exists but is older than origin
                # check if the md5 sum fits
                data = open(targ, "r").readlines()
                _md5 = data[0].rstrip()
                md5_ = self.md5sum(image.path)
                if _md5 == md5_:
                    # md5 sums fit
                    self._return(image, targ, _md5, data[1].rstrip())
                    continue
                self.thumbnail(image, targ)
            else:
                self._return(image, targ)
        
    
    def thumbnail(self, image, targ):
        log("thumbnail (thumbnail) %s" % image.name, 2)
        im = False
        try:
            im = Image.open(image.path)
        except:
            # unable to open origin
            print "Unable to PIL-load image " + image.path
            self.fail(image)
            return
        #try:
        res = "%dx%d" % (im.size[0], im.size[1])
        md5 = self.md5sum(image.path)
        with open(targ, "w") as dataf:
            dataf.write("%s\n%s" % (md5, res))
        im.thumbnail((self.width, self.height), Image.ANTIALIAS)
        if image.metadata and "Exif.Image.Orientation" in image.metadata.exif_keys:
            o = int(image.metadata["Exif.Image.Orientation"].value)
            if o == 2:
                # ##
                #  #
                # ##
                #  #
                im = im.mirror()
            if o == 3:
                #  #
                # ##
                #  #
                # ##
                im = im.mirror()
                im = im.flip()
            if o == 4:
                # #
                # ##
                # #
                # ##
                im = im.flip()
            if o == 5:
                # #####
                # # #
                im = im.rotate(270, expand=True)
                im = im.mirror()
            if o == 6:
                # # #
                # #####
                im = im.rotate(270, expand=True)
            if o == 7:
                #   # #
                # #####
                im = im.mirror()
                im = im.rotate(270, expand=True)
            if o == 8:
                # #####
                #   # #
                im = im.rotate(90, expand=True)
        im.save(targ + ".png")
        self._return(image, targ, md5, res)
        #except:
            #print "Unable to create thumbnail " + self.targ
            #self.fail(self.image)
    
    def md5sum(self, filename):
        md5 = hashlib.md5()
        with open(filename, 'rb') as f: 
            for chnk in iter(lambda: f.read(128 * md5.block_size), b''): 
                md5.update(chnk)
        return md5.hexdigest()
