# Copyright (C) Markus Schmidt <schmidt@boomshop.net>
# License: F4A (free4all) - have a look in the LICENSE file that came with this
# software

import psutil
import gobject
import threading
import Queue

from lib.Thumbnail import Thumbnail
from lib.Utils import Utils

log_level = 1
def log(mesg, level = 0):
    if level <= log_level:
        print mesg
        
class Thumbnailer():
    queue        = [] # list of image instances to thumbnail
    thumbnailers = [] # list of thumbnailers - amout corresponds to the amount of cpu cores
    width        = 0
    height       = 0
    running      = 0
    lock         = threading.Lock()
    tlock        = threading.Lock()
    
    def __init__(self, counter = None, width = 0, height = 0):
        log("__init__ THUMBNAILER", 2)
        # create thumbnail slots
        for i in range(0, psutil.NUM_CPUS or 1):
            self.thumbnailers.append(None)
        self.width   = width
        self.height  = height
        self.counter = counter
        
    def set_size(self, width, height):
        # set the size of the generated thumbnails
        self.width  = width
        self.height = height
    
    def add(self, image):
        log("add (thumbnailer) %s" % image.name, 2)
        # check if image is already in the queue
        self.remove(image)
        # add an image instance to the queue
        self.lock.acquire()
        try:
            self.queue.append(image)
        finally:
            self.lock.release()
        # run the queue
        self.run_queue()
    
    def stop(self, image):
        log("stop (thumbnailer) %s" % image.name)
        self.thumbnailers[image.thumb_id].stop()
        self.thumbnailers[image.thumb_id] = None
        
    def clear(self):
        for i in self.thumbnailers:
            if i:
                i.stop()
        self.lock.acquire()
        self.queue = []
        self.lock.release()
        
    def remove(self, image):
        self.lock.acquire()
        try:
            if image in self.queue:
                self.queue.remove(image)
        finally:
            self.lock.release()
    
    def run_queue(self):
        log("run_queue (thumbnailer) %d" % len(self.queue), 2)
        self.lock.acquire()
        ln = len(self.queue)
        self.lock.release()
        if not self.width or not self.height or not ln:
            # exit if no size was set by now or queue is empty
            self.running = 0
            return
            
        # search for the next avaialbe thumbnail thread
        self.tlock.acquire()
        try:
            for l in range(0, len(self.thumbnailers)):
                if not self.thumbnailers[l]:
                    # we have found a free slot, so go on
                    #image = self.queue.pop(0)
                    #image.thumb_id = l
                    self.running += 1
                    # create and run a new thumbnail instance
                    self.thumbnailers[l] = Thumbnail(self.queue, self.lock, l, self.width, self.height, self.thumb_cb, self.fail_cb, self.done_cb)
        finally:
            self.tlock.release()
            
    def thumb_cb(self, image):
        # callback for the thumbnail processes
        log("thumb_cb (thumbnailer) %s" % image.name, 2)
        self.cleanup(image)
            
    def fail_cb(self, image):
        # do we need to retry? don't know by now.
        log("fail_cb (thumbnailer) %s" % image.name, 1)
        self.cleanup(image)
    
    def done_cb(self, l):
        log("done_cb (thumbnailer) %d" % l, 2)
        self.tlock.acquire()
        self.thumbnailers[l] = None
        self.tlock.release()
        self.running -= 1
        
    def cleanup(self, image):
        log("cleanup (thumbnailer) %s" % image.name, 2)
        self.tlock.acquire()
        ln = len(self.queue)
        self.tlock.release()
        if self.counter:
            self.counter()
        if image and image.folder:
            image.folder.redraw(image)
        if not ln:
            self.running = 0
        else:
            self.run_queue()
