# Copyright (C) Markus Schmidt <schmidt@boomshop.net>
# License: F4A (free4all) - have a look in the LICENSE file that came with this
# software

import os
import re
import datetime
import time

log_level = 2
def log(mesg, level):
    if level >= log_level:
        print mesg
        
class Utils():
    
    def decode_expression(self, exp, image, nums = "", numf = "%d", count = 0):
        d = time.localtime(int(os.path.getctime(image.path)))
        r = {}
        r["f"] = image.name
        r["n"] = r["f"][:image.name.rfind('.')]
        r["e"] = r["f"][image.name.rfind('.') + 1:]
        r["l"] = r["n"].lower()
        r["u"] = r["n"].upper()
        r["d"] = time.strftime(self.lang.datestring, d)
        r["t"] = time.strftime(self.lang.timestring, d)
        r["i"] = ''.join([c for c in r["n"] if c.isdigit()])
        r["c"] = ''.join([c for c in r["n"] if not c.isdigit()])
        if image.metadata and 'Exif.Image.DateTime' in image.metadata.exif_keys:
            r["xd"] = image.metadata['Exif.Image.DateTime'].value.strftime(self.lang.datestring)
            r["xt"] = image.metadata['Exif.Image.DateTime'].value.strftime(self.lang.timestring)
        if nums:
            r[nums] = numf % (int(nums) + count)
        for k, v in r.iteritems():
            exp = re.sub("%" + k, v, exp)
        return exp
        
    def shorten(self, txt, cr, width):
        stop = len(txt) / 2
        start = stop
        (x, y, w, h, dx, dy) = cr.text_extents(txt)
        amount = 3
        while w > width:
            stop = len(txt) / 2 - amount / 2
            start = stop + amount
            (x, y, w, h, dx, dy) = cr.text_extents(txt[:stop] + "..." + txt[start:])
            amount += 1
        if amount > 3:
            return txt[:stop] + "..." + txt[start:]
        else:
            return txt
    
    def bytes2string(self, value):
        """returns a formatted string from bytes (1024)"""
        value = float(value)
        if value >= (1024**4) :
            return self.scaleFloat(value/(1024.0**4)) + " T"
        if value >= (1024**3) :
            return self.scaleFloat(value/(1024.0**3)) + " G"
        if value >= (1024**2) :
            return self.scaleFloat(value/(1024.0**2)) + " M"
        if value >= 1024 :
            return self.scaleFloat(value/1024.0) + " k"
        else : return ("%i" % value) + ""
    
    def scaleFloat(self, value):
        """scales floatings to readable sizes"""
        if value >= 100 :
            return "%d" % round(value)
        if value >= 10 :
            return "%.1f" % value
        else : return "%.2f" % value
    
    def which(self, program):
        def is_exe(fpath):
            return os.path.isfile(fpath) and os.access(fpath, os.X_OK)
    
        fpath, fname = os.path.split(program)
        if fpath:
            if is_exe(program):
                return program
        else:
            for path in os.environ["PATH"].split(os.pathsep):
                path = path.strip('"')
                exe_file = os.path.join(path, program)
                if is_exe(exe_file):
                    return exe_file
        return None
